/*******************************************************************************
 * Copyright (c) 2005 Matthew Hall and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Matthew Hall - initial API and implementation
 ******************************************************************************/
package formslayout;

import java.util.regex.Pattern;

import org.eclipse.swt.SWT;

/**
 * A column spec in a FormLayout. Column specs define the size and weight of
 * columns within a form, and the default horizontal alignment of controls in
 * the column.
 * 
 * @author Matthew
 */
public final class ColSpec extends FormSpec {
	static final int DEFAULT_ALIGN = SWT.LEFT;

	ColSpec(int align, int size, int weight) {
		super(align, size, weight);
	}

	@Override
	int checkAlign(int align) {
		return checkHAlign(align);
	}

	static int checkHAlign(int align) {
		if (align == SWT.LEFT || align == SWT.CENTER || align == SWT.RIGHT
				|| align == SWT.FILL)
			return align;
		if (align == SWT.DEFAULT)
			return DEFAULT_ALIGN;

		throw new IllegalArgumentException("Illegal alignment value: " + align);
	}

	/**
	 * Returns a ColSpec created by parsing the <code>spec</code> argument.
	 * <p>
	 * Format:
	 * 
	 * <pre>
	 * [align:]size[:weight]
	 * 
	 * align  = L | LEFT |               (left alignment)
	 *          C | CENTER |             (center alignment)
	 *          R | RIGHT |              (right alignment)
	 *          F | FILL |               (justify alignment)
	 *          D | DEFAULT              (default alignment)
	 * size   = P | PREF | PREFERRED |   (preferred width)
	 *          D | DEF | DEFAULT |      (preferred width, but will shrink if
	 *                                    space is scarce)
	 *          [#]DLU |                 (# dialog units wide)
	 *          [#]PX                    (# pixels wide)
	 * weight = N | NONE |               (do not grow to fill extra width)
	 *          G | GROW |               (grow to fill extra width)
	 *          G(#) | GROW(#)           (grow to fill extra width using
	 *                                    specific weight)
	 * </pre>
	 * 
	 * The default alignment is LEFT. The
	 * 
	 * <code>weight</code> argument expresses the weight property: NONE
	 * indicates a weight of 0; GROW indicates a weight of 1; and GROW(3)
	 * indicates a weight of 3. The default weight (if <code>weight</code> is
	 * omitted) is 0.
	 * 
	 * <p>
	 * Examples:
	 * 
	 * <pre>
	 * LEFT:DEFAULT:GROW // left-aligned, default width, weight=1
	 * R:100PX:N         // right-aligned, 72 pixels wide, weight=0
	 * right:100px       // identical to previous line
	 * center:20dlu:g(2) // center-aligned, 20 dialog units wide, weight=2
	 * c:pref:none       // center-aligned, preferred width, weight=0
	 * p                 // left-aligned (default), preferred width, weight=0
	 * </pre>
	 * 
	 * @param spec
	 * 		the column spec that will be parsed.
	 * @return a ColSpec created by parsing the <code>spec</code> argument.
	 * @throws IllegalArgumentException
	 * 		if the column spec is incorrectly formatted.
	 */
	protected static ColSpec parse(String spec) {
		String[] matches = spec.split("\\s*:\\s*");

		int align = DEFAULT_ALIGN;
		int size = DEFAULT_SIZE;
		int weight = DEFAULT_WEIGHT;

		if (matches.length == 0)
			throw new IllegalArgumentException("Missing col spec");
		else if (matches.length == 1) {
			// One option: must be size
			size = parseSize(matches[0]);
		} else if (matches.length == 2) {
			// Two possible scenarios:
			// 1. align:size
			// 2. size:weight
			if (isAlign(matches[0]) && isSize(matches[1])) {
				align = parseAlign(matches[0]);
				size = parseSize(matches[1]);
			} else if (isSize(matches[0]) && isWeight(matches[1])) {
				size = parseSize(matches[0]);
				weight = parseWeight(matches[1]);
			}
		} else if (matches.length == 3) {
			align = parseAlign(matches[0]);
			size = parseSize(matches[1]);
			weight = parseWeight(matches[2]);
		}

		return new ColSpec(align, size, weight);
	}

	// Alignment patterns
	private static final Pattern LEFT_ALIGN_PATTERN = Pattern.compile("l|left",
			Pattern.CASE_INSENSITIVE);
	private static final Pattern CENTER_ALIGN_PATTERN = Pattern.compile(
			"c|center", Pattern.CASE_INSENSITIVE);
	private static final Pattern RIGHT_ALIGN_PATTERN = Pattern.compile(
			"r|right", Pattern.CASE_INSENSITIVE);
	private static final Pattern DEFAULT_ALIGN_PATTERN = Pattern.compile(
			"d|default", Pattern.CASE_INSENSITIVE);
	private static final Pattern FILL_ALIGN_PATTERN = Pattern.compile("f|fill",
			Pattern.CASE_INSENSITIVE);
	private static final Pattern ANY_ALIGN_PATTERN = Pattern.compile(
			"l|left|c|center|r|right|d|def|default|f|fill",
			Pattern.CASE_INSENSITIVE);

	private static boolean isAlign(String align) {
		return ANY_ALIGN_PATTERN.matcher(align).matches();
	}

	static int parseAlign(String align) {
		if (LEFT_ALIGN_PATTERN.matcher(align).matches())
			return SWT.LEFT;
		else if (CENTER_ALIGN_PATTERN.matcher(align).matches())
			return SWT.CENTER;
		else if (RIGHT_ALIGN_PATTERN.matcher(align).matches())
			return SWT.RIGHT;
		else if (DEFAULT_ALIGN_PATTERN.matcher(align).matches())
			return SWT.DEFAULT;
		else if (FILL_ALIGN_PATTERN.matcher(align).matches())
			return SWT.FILL;
		else
			throw new IllegalArgumentException("Invalid alignment: \"" + align
					+ "\"");
	}

	ColSpec copy() {
		return new ColSpec(align, size, weight);
	}
}
