/*******************************************************************************
 * Copyright (c) 2005 Matthew Hall and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Matthew Hall - initial API and implementation
 ******************************************************************************/
package formslayout;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Control;

/**
 * Layout data class for FormLayout.
 * 
 * @author Matthew
 */
public class FormData {
	final int x;
	final int y;
	final int cols;
	final int rows;
	final int hAlign;
	final int vAlign;

	// Cached layout information
	int cacheWidth = -1;
	int cacheHeight = -1;

	int defaultWidth = -1;
	int defaultHeight = -1;

	int currentWHint = -1;
	int currentHHint = -1;
	int currentWidth = -1;
	int currentHeight = -1;

	private FormData(int x, int y, int cols, int rows, int hAlign, int vAlign) {
		// sanity check
		if (x < 0)
			throw new IllegalArgumentException("x < 0");
		if (y < 0)
			throw new IllegalArgumentException("y < 0");
		if (cols < 1)
			throw new IllegalArgumentException("cols < 1");
		if (rows < 1)
			throw new IllegalArgumentException("rows < 1");
		ColSpec.checkHAlign(hAlign);
		RowSpec.checkVAlign(vAlign);

		this.x = x;
		this.y = y;
		this.cols = cols;
		this.rows = rows;
		this.hAlign = hAlign;
		this.vAlign = vAlign;
	}

	void flushCache() {
		cacheWidth = cacheHeight = -1;
		defaultWidth = defaultHeight = -1;
		currentWidth = currentHeight = -1;
	}

	void computeSize(Control control, int wHint, int hHint, boolean flushCache) {
		// The flushCache argument is intended for calls to control.computeSize(
		// ),
		// not for this FormData.

		// Special case: calculate the control size with SWT.DEFAULT width and
		// height hints. Store this default size in separate fields from the
		// other cached sizes.
		if (wHint == SWT.DEFAULT && hHint == SWT.DEFAULT) {
			if (defaultWidth == -1 || defaultHeight == -1) {
				Point size = control.computeSize(wHint, hHint, flushCache);
				defaultWidth = size.x;
				defaultHeight = size.y;
			}
			cacheWidth = defaultWidth;
			cacheHeight = defaultHeight;
			return;
		}

		if (currentWidth == -1 || currentHeight == -1 || wHint != currentWHint
				|| hHint != currentHHint) {
			Point size = control.computeSize(wHint, hHint, flushCache);
			currentWHint = wHint;
			currentHHint = hHint;
			currentWidth = size.x;
			currentHeight = size.y;
		}
		cacheWidth = currentWidth;
		cacheHeight = currentHeight;
	}

	int computeWidth(Control control, int hHint, boolean flushCache) {
		computeSize(control, SWT.DEFAULT, hHint, flushCache);
		return cacheWidth;
	}

	int computeHeight(Control control, int wHint, boolean flushCache) {
		computeSize(control, wHint, SWT.DEFAULT, flushCache);
		return cacheHeight;
	}

	private static int[] parseAlign(String align) {
		String[] aligns = align.split("\\s*,\\s*");
		if (aligns.length != 2)
			throw new IllegalArgumentException(
					"Alignment must have two comma-separated alignment attributes.");

		int[] result = new int[2];
		result[0] = ColSpec.parseAlign(aligns[0]);
		result[1] = RowSpec.parseAlign(aligns[1]);

		return result;
	}

	/**
	 * Returns a FormData which places the control in cell (x, y) in the form,
	 * with default alignments.
	 * 
	 * @param x
	 * 		the column index of the cell the control is placed in.
	 * @param y
	 * 		the row index of the cell the control is placed in.
	 * @return a FormData placing the control in the given cell, using the
	 * 	default row and column alignments.
	 */
	public static FormData xy(int x, int y) {
		return xy(x, y, SWT.DEFAULT, SWT.DEFAULT);
	}

	/**
	 * Returns a FormData placing the control in cell (x, y) in the form, with
	 * the given alignments.
	 * 
	 * @param x
	 * 		the column index of the cell the control is placed in.
	 * @param y
	 * 		the column index of the cell the control is placed in.
	 * @param hAlign
	 * 		the horizontal alignment of the control within the cell. Must be one
	 * 		of SWT.LEFT, SWT.CENTER, SWT.RIGHT, SWT.FILL, or SWT.DEFAULT.
	 * @param vAlign
	 * 		the vertical alignment of the control within the cell. Must be one
	 * 		of SWT.TOP, SWT.CENTER, SWT.BOTTOM, SWT.FILL, or SWT.DEFAULT.
	 * @return a FormData placing the control in the given cell, using the given
	 * 	horizontal and vertical alignments.
	 */
	public static FormData xy(int x, int y, int hAlign, int vAlign) {
		return new FormData(x, y, 1, 1, hAlign, vAlign);
	}

	/**
	 * Returns a FormData placing the control in cell (x, y) in the form, with
	 * the given alignments.
	 * 
	 * @param x
	 * 		the column index of the cell the control is placed in.
	 * @param y
	 * 		the column index of the cell the control is placed in.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return a FormData placing the control in the given cell, using the given
	 * 	alignments.
	 */
	public static FormData xy(int x, int y, String align) {
		int[] aligns = parseAlign(align);
		return xy(x, y, aligns[0], aligns[1]);
	}

	/**
	 * Returns a FormData which places the control in a cell span starting at
	 * (x, y) in the form, spanning <code>w</code> cells horizontally, and using
	 * the default row and column alignments of the starting cell.
	 * 
	 * @param x
	 * 		the column index of the cell the control is placed in.
	 * @param y
	 * 		the row index of the cell the control is placed in.
	 * @param w
	 * 		the horizontal cell span. Must be 1 or more.
	 * @return a FormData placing the control in the given cell span, using the
	 * 	default alignments.
	 */
	public static FormData xyw(int x, int y, int w) {
		return xyw(x, y, w, SWT.FILL, SWT.DEFAULT);
	}

	/**
	 * Returns a FormData which places the control in a cell span starting at
	 * (x, y) in the form, spanning <code>w</code> cells horizontally, using the
	 * given alignments.
	 * 
	 * @param x
	 * 		the column index of the cell the control is placed in.
	 * @param y
	 * 		the row index of the cell the control is placed in.
	 * @param w
	 * 		the horizontal cell span. Must be 1 or more.
	 * @param hAlign
	 * 		the horizontal alignment of the control within the cell. Must be one
	 * 		of SWT.LEFT, SWT.CENTER, SWT.RIGHT, SWT.FILL, or SWT.DEFAULT.
	 * @param vAlign
	 * 		the vertical alignment of the control within the cell. Must be one
	 * 		of SWT.TOP, SWT.CENTER, SWT.BOTTOM, SWT.FILL, or SWT.DEFAULT.
	 * @return a FormData placing the control in the given cell span, using the
	 * 	given alignments.
	 */
	public static FormData xyw(int x, int y, int w, int hAlign, int vAlign) {
		return new FormData(x, y, w, 1, hAlign, vAlign);
	}

	/**
	 * Returns a FormData which places the control in a cell span starting at
	 * (x, y) in the form, spanning <code>w</code> cells horizontally, using the
	 * given alignments.
	 * 
	 * @param x
	 * 		the column index of the cell the control is placed in.
	 * @param y
	 * 		the row index of the cell the control is placed in.
	 * @param w
	 * 		the horizontal cell span. Must be 1 or more.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return a FormData placing the control in the given cell span, using the
	 * 	given alignments.
	 */
	public static FormData xyw(int x, int y, int w, String align) {
		int[] aligns = parseAlign(align);
		return xyw(x, y, w, aligns[0], aligns[1]);
	}

	/**
	 * Returns a FormData which places the control in a cell span starting at
	 * (x, y) in the form, spanning <code>w</code> cells horizontally, spanning
	 * <code>h</code> cells vertically, and using the default row and column
	 * alignments of the starting cell.
	 * 
	 * @param x
	 * 		the column index of the cell the control is placed in.
	 * @param y
	 * 		the row index of the cell the control is placed in.
	 * @param w
	 * 		the horizontal cell span. Must be 1 or more.
	 * @param h
	 * 		the vertical cell span. Must be 1 or more.
	 * @return a FormData placing the control in the given cell span, using the
	 * 	default alignments.
	 */
	public static FormData xywh(int x, int y, int w, int h) {
		return xywh(x, y, w, h, SWT.FILL, SWT.FILL);
	}

	/**
	 * Returns a FormData which places the control in a cell span starting at
	 * (x, y) in the form, spanning <code>w</code> cells horizontally, spanning
	 * <code>h</code> cells vertically, and using the given horizontal and
	 * vertical alignments.
	 * 
	 * @param x
	 * 		the column index of the cell the control is placed in.
	 * @param y
	 * 		the row index of the cell the control is placed in.
	 * @param w
	 * 		the horizontal cell span. Must be 1 or more.
	 * @param h
	 * 		the vertical cell span. Must be 1 or more.
	 * @param hAlign
	 * 		the horizontal alignment of the control within the cell. Must be one
	 * 		of SWT.LEFT, SWT.CENTER, SWT.RIGHT, SWT.FILL, or SWT.DEFAULT.
	 * @param vAlign
	 * 		the vertical alignment of the control within the cell. Must be one
	 * 		of SWT.TOP, SWT.CENTER, SWT.BOTTOM, SWT.FILL, or SWT.DEFAULT.
	 * @return a FormData placing the control in the given cell span, using the
	 * 	given alignments.
	 */
	public static FormData xywh(int x, int y, int w, int h, int hAlign,
			int vAlign) {
		return new FormData(x, y, w, h, hAlign, vAlign);
	}

	/**
	 * Returns a FormData which places the control in a cell span starting at
	 * (x, y) in the form, spanning <code>w</code> cells horizontally, spanning
	 * <code>h</code> cells vertically, and using the given horizontal and
	 * vertical alignments.
	 * 
	 * @param x
	 * 		the column index of the cell the control is placed in.
	 * @param y
	 * 		the row index of the cell the control is placed in.
	 * @param w
	 * 		the horizontal cell span. Must be 1 or more.
	 * @param h
	 * 		the vertical cell span. Must be 1 or more.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return a FormData placing the control in the given cell span, using the
	 * 	given alignments.
	 */
	public static FormData xywh(int x, int y, int w, int h, String align) {
		int[] aligns = parseAlign(align);
		return xywh(x, y, w, h, aligns[0], aligns[1]);
	}

	@Override
	public final String toString() {
		StringBuffer buf = new StringBuffer(26);
		buf.append("FormData [");
		buf.append(x);
		buf.append(", ");
		buf.append(y);
		buf.append(", ");
		buf.append(cols);
		buf.append(", ");
		buf.append(rows);
		buf.append("]");
		return buf.toString();
	}
}
