/*******************************************************************************
 * Copyright (c) 2005 Matthew Hall and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Matthew Hall - initial API and implementation
 ******************************************************************************/
package formslayout;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Layout;

/**
 * A layout class for positioning controls in a tabular form.
 * 
 * @author Matthew
 */
public final class FormLayout extends Layout {
	static final int PREFERRED = -2;

	final List<RowSpec> rows;
	private int[][] columnGroups = new int[0][];

	final List<ColSpec> cols;
	private int[][] rowGroups = new int[0][];

	private int margin;

	/**
	 * Constructs a FormLayout with the given column specs. Rows may be added
	 * through the addRow(...) methods.
	 * 
	 * @param colspec
	 * 		a comma-delimited string of parseable column specs.
	 * @see ColSpec#parse(String)
	 */
	public FormLayout(String colspec) {
		this(colspec, "", 0);
	}

	/**
	 * Constructs a FormLayout with the given column specs and margins. Rows may
	 * be added through the addRow(...) methods.
	 * 
	 * @param colspec
	 * 		a comma-delimited string of parseable column specs.
	 * @param margin
	 * 		the margin, in pixels.
	 * @see ColSpec#parse(String)
	 * @see RowSpec#parse(String)
	 */
	public FormLayout(String colspec, int margin) {
		this(colspec, "", margin);
	}

	/**
	 * Constructs a FormLayout with the given column and row specs.
	 * 
	 * @param colspec
	 * 		a comma-delimited string of parseable column specs.
	 * @param rowspec
	 * 		a comma-delimited string of parseable row specs.
	 * @see ColSpec#parse(String)
	 * @see RowSpec#parse(String)
	 */
	public FormLayout(String colspec, String rowspec) {
		this(colspec, rowspec, 0);
	}

	/**
	 * Constructs a FormLayout with the given column specs, row specs, and
	 * margins.
	 * 
	 * @param colspec
	 * 		a comma-delimited string of parseable column specs.
	 * @param rowspec
	 * 		a comma-delimited string of parseable row specs.
	 * @param margin
	 * 		the margin, in pixels.
	 * @see ColSpec#parse(String)
	 * @see RowSpec#parse(String)
	 */
	public FormLayout(String colspec, String rowspec, int margin) {
		String[] cols = colspec.split("\\s*,\\s*");
		this.cols = new ArrayList<ColSpec>(cols.length);
		for (int i = 0; i < cols.length; i++)
			addColumn(cols[i]);

		String[] rows = rowspec.split("\\s*,\\s*");
		this.rows = new ArrayList<RowSpec>(rows.length);
		for (int i = 0; i < rows.length; i++)
			addRow(rows[i]);

		this.margin = Math.max(0, margin);
	}

	/**
	 * Adds the given column spec at the right of the layout.
	 * 
	 * @param colspec
	 * 		a comma-delimited string of parseable column specs.
	 * @see ColSpec#parse(String)
	 */
	public void addColumn(String colspec) {
		addColumn(cols.size(), colspec);
	}

	/**
	 * Adds the given column spec at the given position in the layout.
	 * 
	 * @param index
	 * 		the index where the column should be added.
	 * @param colspec
	 * 		a comma-delimited string of parseable column specs.
	 * @see ColSpec#parse(String)
	 */
	public void addColumn(int index, String colspec) {
		cols.add(index, ColSpec.parse(colspec));
	}

	/**
	 * Removes the column spec at the given position in the layout.
	 * 
	 * @param index
	 * 		the index of the column to remove.
	 */
	public void removeColumn(int index) {
		cols.remove(index);
	}

	/**
	 * Adds the given row spec at the bottom of the layout.
	 * 
	 * @param rowspec
	 * 		a comma-delimited string of parseable row specs.
	 * @see RowSpec#parse(String)
	 */
	public void addRow(String rowspec) {
		addRow(rows.size(), rowspec);
	}

	/**
	 * Adds the given row spec at the given position in the layout.
	 * 
	 * @param index
	 * 		the index where the row should be added.
	 * @param rowspec
	 * 		a comma-delimited string of parseable row specs.
	 * @see RowSpec#parse(String)
	 */
	public void addRow(int index, String rowspec) {
		rows.add(index, RowSpec.parse(rowspec));
	}

	/**
	 * Removes the row spec at the given position in the layout.
	 * 
	 * @param index
	 * 		the index of the row to remove.
	 */
	public void removeRow(int index) {
		rows.remove(index);
	}

	/**
	 * Returns a two-dimensional array of the columns that will be grouped.
	 * 
	 * @return a two-dimensional array of the columns that will be grouped.
	 */
	public int[][] getColumnGroups() {
		return columnGroups;
	}

	/**
	 * Sets the column groups. Columns in the same group are given the same size
	 * in the form.
	 * 
	 * @param columnGroups
	 * 		a two-dimensional array wherein the each element in the top-level
	 * 		array is a column group, and each element in the sub-array is the
	 * 		index of a column in the group.
	 */
	public void setColumnGroups(int[][] columnGroups) {
		// sanity check
		if (columnGroups == null)
			throw new NullPointerException();

		for (int[] group : columnGroups) {
			if (group == null)
				throw new NullPointerException();

			for (int col : group)
				if (col < 0)
					throw new IndexOutOfBoundsException(
							"Column out of bounds: " + col + " < 0");
				else if (col >= cols.size())
					throw new IndexOutOfBoundsException(
							"Column out of bounds: " + col + " >= "
									+ cols.size());
		}

		this.columnGroups = columnGroups;
	}

	/**
	 * Returns a two-dimensional array of the rows that will be grouped.
	 * 
	 * @return a two-dimensional array of the rows that will be grouped.
	 */
	public int[][] getRowGroups() {
		return rowGroups;
	}

	/**
	 * Sets the row groups. Rows in the same group are given the same size in
	 * the form.
	 * 
	 * @param rowGroups
	 * 		a two-dimensional array wherein the each element in the top-level
	 * 		array is a row group, and each element in the sub-array is the index
	 * 		of a row in the group.
	 */
	public void setRowGroups(int[][] rowGroups) {
		// sanity check
		if (rowGroups == null)
			throw new NullPointerException();

		for (int[] group : rowGroups) {
			if (group == null)
				throw new NullPointerException();

			for (int row : group)
				if (row < 0)
					throw new IndexOutOfBoundsException("Row out of bounds: "
							+ row + " < 0");
				else if (row >= rows.size())
					throw new IndexOutOfBoundsException("Row out of bounds: "
							+ row + " >= " + rows.size());
		}

		this.rowGroups = rowGroups;
	}

	/**
	 * Make sure the layout data fits within the form grid.
	 * 
	 * @param data
	 * 		the FormData to check.
	 * @throws IndexOutOfBoundsException
	 * 		if the cellspan described by the argument is outside the boundaries
	 * 		of this FormLayout.
	 */
	public void checkFormData(FormData data) {
		if (data.x >= cols.size())
			throw new IndexOutOfBoundsException("x = " + data.x + " >= "
					+ cols.size() + " (# of columns)");
		if (data.x + data.cols > cols.size())
			throw new IndexOutOfBoundsException("x + w = "
					+ (data.x + data.cols) + " > " + cols.size()
					+ " (# of columns)");

		if (data.y >= rows.size())
			throw new IndexOutOfBoundsException("y = " + data.y + " >= "
					+ rows.size() + " (# of rows)");
		if (data.y + data.rows > rows.size())
			throw new IndexOutOfBoundsException("y + h = "
					+ (data.y + data.rows) + " > " + rows.size()
					+ " (# of rows)");
	}

	@Override
	protected Point computeSize(Composite composite, int wHint, int hHint,
			boolean flushCache) {
		// Calculate column widths.
		int[] colSizes;
		int[] rowSizes;

		// Width hint supersedes height hint if wHint != SWT.DEFAULT
		if (wHint != SWT.DEFAULT || hHint == SWT.DEFAULT) {
			colSizes = computeColSizes(composite, wHint, null, flushCache);
			rowSizes = computeRowSizes(composite, hHint, colSizes, flushCache);
		} else {
			rowSizes = computeRowSizes(composite, hHint, null, flushCache);
			colSizes = computeColSizes(composite, wHint, rowSizes, flushCache);
		}

		int width = 2 * margin;
		int height = 2 * margin;
		for (int col : colSizes)
			width += col;
		for (int row : rowSizes)
			height += row;

		return new Point(width, height);
	}

	@Override
	protected boolean flushCache(Control control) {
		Object data = control.getLayoutData();
		if (data != null)
			((FormData) data).flushCache();
		return true;
	}

	private int[] computeColSizes(Composite composite, int wHint,
			int[] rowSizes, boolean flushCache) {

		Control[] children = composite.getChildren();
		int[] colSizes = new int[cols.size()];
		int width = 0;
		int weight = 0;

		if (wHint != SWT.DEFAULT)
			wHint -= Math.min(wHint, 2 * margin);

		// Calculate column sizes based on controls
		for (int colI = 0; colI < colSizes.length; colI++) {
			ColSpec col = cols.get(colI);
			if (col.size == SWT.DEFAULT || col.size == PREFERRED) {
				// calculate size
				int size = 0;
				for (int childI = 0; childI < children.length; childI++) {
					Control child = children[childI];

					// Skip controls if handled in a previous iteration
					if (child == null)
						continue;

					// Ignore invisible controls
					if (!child.getVisible()) {
						children[childI] = null; // Skip in future iterations.
						continue;
					}

					FormData data = (FormData) child.getLayoutData();

					// Cannot layout if there is no FormData
					if (data == null) {
						children[childI] = null; // Skip control in future
						// iterations.
						continue;
					}

					// Check for FormData out of bounds.
					if (data.x + data.cols > cols.size()
							|| data.y + data.rows > rows.size()) {
						children[childI] = null;
						continue;
					}

					if (flushCache)
						data.flushCache();

					// Ignore multi-column controls in this loop.
					if (data.cols != 1)
						continue;

					// If the current control is in the column we care about,
					// update size
					if (data.x == colI) {
						int hHint = rowSizes == null ? SWT.DEFAULT : sum(
								rowSizes, data.y, data.rows);

						size = Math.max(size, data.computeWidth(child, hHint,
								flushCache));
						children[childI] = null; // skip control in future
						// iterations.
					}
				}
				colSizes[colI] = size;
			} else {
				colSizes[colI] = col.size;
			}
			width += colSizes[colI];
			weight += col.weight;
		}

		// Equalize columns in each column group
		for (int[] group : columnGroups) {
			// Find the widest column
			int widest = 0;
			for (int col : group)
				widest = Math.max(widest, colSizes[col]);

			// Make each column in group the same width
			for (int col : group) {
				width += (widest - colSizes[col]);
				colSizes[col] = widest;
			}
		}

		// Now check the multi-column controls, expanding columns where
		// necessary.
		for (Control child : children) {
			if (child == null)
				continue;

			FormData data = (FormData) child.getLayoutData();

			int hHint = rowSizes == null ? SWT.DEFAULT : sum(rowSizes, data.y,
					data.rows);
			int controlWidth = data.computeWidth(child, hHint, flushCache);
			int spanWidth = sum(colSizes, data.x, data.cols);

			if (controlWidth > spanWidth) {
				int delta = controlWidth - spanWidth;
				growForSpanningElement(cols, colSizes, delta, data.x, data.cols);
				width += delta;
			}
		}

		// If wHint is an explicit size that is less than the preferred width,
		// shrink DEFAULT sized columns to fit within wHint.
		if (wHint != SWT.DEFAULT && width > wHint) {
			shrinkDefaultSizeElements(cols, colSizes, width - wHint, 0,
					colSizes.length);

			// Recalculate width to see if the default size columns were able to
			// shrink enough
			width = sum(colSizes, 0, colSizes.length);

			// If the columns are still too wide, completely collapse the
			// columns one at a time starting
			// at the right side, until the layout fits.
			if (width > wHint)
				for (int i = colSizes.length - 1; i >= 0 && width > wHint; i--) {
					int subtract = Math.min(colSizes[i], width - wHint);
					colSizes[i] -= subtract;
					width -= subtract;
				}
		}

		// If wHint is an explicit size that is more than the preferred width,
		// distribute the extra width between columns according to weight.
		else if (wHint > width && weight > 0)
			growWeightedElements(cols, colSizes, wHint - width, weight);

		return colSizes;
	}

	private int[] computeRowSizes(Composite composite, int hHint,
			int[] colSizes, boolean flushCache) {

		Control[] children = composite.getChildren();
		int[] rowSizes = new int[rows.size()];
		int height = 0;
		int weight = 0;

		if (hHint != SWT.DEFAULT)
			hHint -= Math.min(hHint, 2 * margin);

		// Calculate row sizes based on controls
		for (int rowI = 0; rowI < rowSizes.length; rowI++) {
			RowSpec row = rows.get(rowI);
			if (row.size == SWT.DEFAULT || row.size == PREFERRED) {
				// calculate size
				int size = 0;
				for (int childI = 0; childI < children.length; childI++) {
					Control child = children[childI];

					// skip controls if handled in a previous iteration
					if (child == null)
						continue;

					// Ignore invisible controls
					if (!child.getVisible()) {
						children[childI] = null;
						continue;
					}

					FormData data = (FormData) child.getLayoutData();

					// Cannot layout if there is no FormData
					if (data == null) {
						children[childI] = null; // skip control in future
						// iterations.
						continue;
					}

					// Check for FormData out of bounds.
					if (data.x + data.cols > cols.size()
							|| data.y + data.rows > rows.size()) {
						children[childI] = null;
						continue;
					}

					if (flushCache)
						data.flushCache();

					// Only single-row controls affect the row size.
					if (data.rows != 1)
						continue;

					// If the current control is in the row we care about,
					// update size
					if (data.y == rowI) {
						int wHint = colSizes == null ? SWT.DEFAULT : sum(
								colSizes, data.x, data.cols);

						size = Math.max(size, data.computeHeight(child, wHint,
								flushCache));
						children[childI] = null; // skip control in future
						// iterations.
					}
				}
				rowSizes[rowI] = size;
			} else {
				rowSizes[rowI] = row.size;
			}
			height += rowSizes[rowI];
			weight += row.weight;
		}

		// Equalize rows in each row group
		for (int[] group : rowGroups) {
			// Find the tallest row
			int tallest = 0;
			for (int row : group)
				tallest = Math.max(tallest, rowSizes[row]);

			// Make each row in the group the same height
			for (int row : group) {
				height += (tallest - rowSizes[row]);
				rowSizes[row] = tallest;
			}
		}

		// Now check the multi-row controls, expanding rows where necessary.
		for (Control child : children) {
			if (child == null)
				continue;

			FormData data = (FormData) child.getLayoutData();

			int wHint = colSizes == null ? SWT.DEFAULT : sum(colSizes, data.x,
					data.cols);
			int controlHeight = data.computeHeight(child, wHint, flushCache);
			int spanHeight = sum(rowSizes, data.y, data.rows);

			if (controlHeight > spanHeight) {
				int delta = controlHeight - spanHeight;
				growForSpanningElement(rows, rowSizes, delta, data.y, data.rows);
				height += delta;
			}
		}

		// If hHint is an explicit size that is less than the preferred height,
		// shrink DEFAULT sized rows to fit within hHint.
		if (hHint != SWT.DEFAULT && height > hHint) {
			shrinkDefaultSizeElements(rows, rowSizes, height - hHint, 0,
					rowSizes.length);

			// Recalculate height to see if the default size rows were able to
			// shrink enough
			height = sum(rowSizes, 0, rowSizes.length);

			// If the rows are still too tall, completely collapse the row one
			// at a time starting
			// at the bottom, until the layout fits.
			if (height > hHint)
				for (int i = rowSizes.length - 1; i >= 0 && height > hHint; i--) {
					int subtract = Math.min(rowSizes[i], height - hHint);
					rowSizes[i] -= subtract;
					height -= subtract;
				}
		}

		// If hHint is an explicit size that is more than the preferred height,
		// distribute the extra height between rows according to weight.
		else if (hHint > height && weight > 0)
			growWeightedElements(rows, rowSizes, hHint - height, weight);

		return rowSizes;
	}

	private static int sum(int[] sizes, int offset, int count) {
		int result = 0;
		for (int index = offset; index < offset + count; index++)
			result += sizes[index];
		return result;
	}

	private static void growForSpanningElement(List<? extends FormSpec> specs,
			int[] sizes, int delta, int offset, int count) {
		List<Integer> indices = new ArrayList<Integer>();

		// Search for default size elements first.
		for (int i = offset; i < offset + count; i++)
			if (specs.get(i).size == SWT.DEFAULT)
				indices.add(i);

		// Search for weighted elements next.
		if (indices.size() == 0)
			for (int i = offset; i < offset + count; i++)
				if (specs.get(i).weight > 0)
					indices.add(i);

		// Search for preferred size elements next.
		if (indices.size() == 0)
			for (int i = offset; i < offset + count; i++)
				if (specs.get(i).size == PREFERRED)
					indices.add(i);

		for (Iterator<Integer> iter = indices.iterator(); iter.hasNext();) {
			int index = iter.next();
			int addWidth = delta / indices.size();

			sizes[index] += addWidth;
			delta -= addWidth;
			iter.remove();
		}
	}

	private static void growWeightedElements(List<? extends FormSpec> specs,
			int[] sizes, int delta, int weight) {
		for (int i = 0; i < sizes.length; i++) {
			FormSpec spec = specs.get(i);
			if (spec.weight > 0) {
				// Use (columnWeight / totalWeight) of the extra width
				int addSize = delta * spec.weight / weight;
				delta -= addSize;
				weight -= spec.weight;
				sizes[i] += addSize;
			}
		}
	}

	private static void shrinkDefaultSizeElements(
			List<? extends FormSpec> specs, int[] sizes, int delta, int offset,
			int count) {
		int defaultSize = 0;
		for (int i = 0; i < sizes.length; i++)
			if (specs.get(i).size == SWT.DEFAULT)
				defaultSize += sizes[i];

		if (defaultSize < delta) {
			// The size we need to take out is more than the size available in
			// the default columns.
			// Best we can do is reduce all default size elements to 0.
			for (int i = 0; i < sizes.length; i++)
				if (specs.get(i).size == SWT.DEFAULT)
					sizes[i] = 0;
		} else {
			// Scale down each element proportionate to its current size.
			for (int i = 0; i < sizes.length; i++)
				if (specs.get(i).size == SWT.DEFAULT) {
					int subtractSize = delta * sizes[i] / defaultSize;
					defaultSize -= sizes[i];
					delta -= subtractSize;
					sizes[i] -= subtractSize;
					if (defaultSize <= 0)
						break;
				}
		}
	}

	@Override
	protected void layout(Composite composite, boolean flushCache) {
		Rectangle rect = composite.getClientArea();
		int[] colSizes = computeColSizes(composite, rect.width, null,
				flushCache);
		int[] rowSizes = computeRowSizes(composite, rect.height, colSizes,
				flushCache);

		// determine the x coordinate where each column starts.
		int cursor = margin;
		int[] colStarts = new int[colSizes.length];
		for (int i = 0; i < colStarts.length; i++) {
			colStarts[i] = cursor;
			cursor += colSizes[i];
		}

		// determine the y coordinate where each row starts.
		cursor = margin;
		int[] rowStarts = new int[rowSizes.length];
		for (int i = 0; i < rowStarts.length; i++) {
			rowStarts[i] = cursor;
			cursor += rowSizes[i];
		}

		for (Control child : composite.getChildren()) {
			FormData data = (FormData) child.getLayoutData();

			// Ignore control if no layout data, or the child is not visible.
			if (data == null) {
				child.setBounds(0, 0, 0, 0);
				continue;
			}

			// If the form data is out of bounds, hide the control.
			if (data.x + data.cols > cols.size()
					|| data.y + data.rows > rows.size()) {
				// Print an error messages if the control is visible.
				if (child.getVisible())
					System.err.println("FormData out of bounds: " + data);

				child.setBounds(0, 0, 0, 0);
				continue;
			}

			ColSpec colSpec = cols.get(data.x);
			RowSpec rowSpec = rows.get(data.y);

			// Obtain dimensions of cell span the the control occupies
			int x = colStarts[data.x];
			int y = rowStarts[data.y];

			int width = 0;
			for (int i = data.x; i < data.x + data.cols; i++)
				width += colSizes[i];

			int height = 0;
			for (int i = data.y; i < data.y + data.rows; i++)
				height += rowSizes[i];

			// Determine alignment for this control
			int hAlign = resolveAlign(colSpec.align, data.hAlign);
			if (hAlign == SWT.DEFAULT)
				hAlign = ColSpec.DEFAULT_ALIGN;

			int vAlign = resolveAlign(rowSpec.align, data.vAlign);
			if (vAlign == SWT.DEFAULT)
				vAlign = RowSpec.DEFAULT_ALIGN;

			// Get the preferred size of the control
			Point preferredSize = child.computeSize(SWT.DEFAULT, SWT.DEFAULT);

			// If preferred height doesn't fit into the cell span, recalculate
			// with
			// height hint.
			if (preferredSize.y > height)
				preferredSize = child.computeSize(SWT.DEFAULT, height);

			// If preferred width doesn't fit into the cell span, recalculate
			// with
			// width hint. Note that we check width second so that it overrides
			// height in case of a conflict.
			if (preferredSize.x > width)
				preferredSize = child.computeSize(width, SWT.DEFAULT);

			// Determine the control's location and size within the cell span,
			// using the horizontal and vertical alignment properties.
			if (hAlign == SWT.LEFT) {
				width = Math.min(preferredSize.x, width);
			} else if (hAlign == SWT.CENTER) {
				int controlWidth = Math.min(preferredSize.x, width);
				x = x + (width - controlWidth) / 2;
				width = controlWidth;
			} else if (hAlign == SWT.RIGHT) {
				int controlWidth = Math.min(preferredSize.x, width);
				x = x + width - controlWidth;
				width = controlWidth;
			} else if (hAlign == SWT.FILL) {
				// Leave x and width as is
			} else {
				throw new IllegalStateException(
						"Illegal alignment for control " + child + " (hAlign="
								+ hAlign + ")");
			}

			if (vAlign == SWT.TOP) {
				height = Math.min(preferredSize.y, height);
			} else if (vAlign == SWT.CENTER) {
				int controlHeight = Math.min(preferredSize.y, height);
				y = y + (height - controlHeight) / 2;
				height = controlHeight;
			} else if (vAlign == SWT.BOTTOM) {
				int controlHeight = Math.min(preferredSize.y, height);
				y = y + height - controlHeight;
				height = controlHeight;
			} else if (vAlign == SWT.FILL) {
				// Leave y and height as is
			} else {
				throw new IllegalStateException(
						"Illegal alignment for control " + child + " (vAlign="
								+ vAlign + ")");
			}

			child.setBounds(x, y, width, height);
		}
	}

	/**
	 * Choose between the control alignment or the default alignment of the row
	 * or column. If the control alignment is SWT.DEFAULT, the default alignment
	 * of the row or column is used. Otherwise, the control's alignment is used.
	 */
	private static int resolveAlign(int defaultAlign, int controlAlign) {
		return controlAlign == SWT.DEFAULT ? defaultAlign : controlAlign;
	}
}
