/*******************************************************************************
 * Copyright (c) 2005 Matthew Hall and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Matthew Hall - initial API and implementation
 ******************************************************************************/
package formslayout;

import java.util.regex.Pattern;

import org.eclipse.swt.SWT;

/**
 * A row spec in a FormLayout. Row specs define the size and weight of rows
 * within a form, and the default vertical alignment of controls in the row.
 * 
 * @author Matthew
 */
public final class RowSpec extends FormSpec {
	static final int DEFAULT_ALIGN = SWT.CENTER;

	RowSpec(int align, int size, int weight) {
		super(align, size, weight);
	}

	@Override
	int checkAlign(int align) {
		return checkVAlign(align);
	}

	static int checkVAlign(int align) {
		if (align == SWT.TOP || align == SWT.CENTER || align == SWT.BOTTOM
				|| align == SWT.FILL)
			return align;
		if (align == SWT.DEFAULT)
			return DEFAULT_ALIGN;

		throw new IllegalArgumentException("Illegal alignment value: " + align);
	}

	/**
	 * Returns a RowSpec created by parsing the <code>spec</code> argument.
	 * <p>
	 * Format:
	 * 
	 * <pre>
	 * [align:]size[:weight]
	 * 
	 * align  = T | TOP |                (top alignment)
	 *          C | CENTER |             (center alignment)
	 *          B | BOTTOM |             (bottom alignment)
	 *          F | FILL |               (justify alignment)
	 *          D | DEFAULT              (default alignment)
	 * size   = P | PREF | PREFERRED |   (preferred height)
	 *          D | DEF | DEFAULT |      (like preferred, but will shrink if
	 *                                    space is scarce)
	 *          [#]DLU |                 (# dialog units tall)
	 *          [#]PX                    (# pixels tall)
	 * weight = N | NONE |               (do not grow to fill extra height)
	 *          G | GROW |               (grow to fill extra height)
	 *          G(#) | GROW(#)           (grow to fill extra height using
	 *                                    specific weight)
	 * </pre>
	 * 
	 * The default alignment is TOP. The
	 * 
	 * <code>weight</code> argument expresses the weight property: NONE
	 * indicates a weight of 0; GROW indicates a weight of 1; and GROW(3)
	 * indicates a weight of 3. The default weight (if <code>weight</code> is
	 * omitted) is 0.
	 * 
	 * <p>
	 * Examples:
	 * 
	 * <pre>
	 * TOP:DEFAULT:GROW  // top-aligned, default height, weight=1
	 * B:100PX:N         // bottom-aligned, 72 pixels tall, weight=0
	 * bottom:100px      // identical to previous line
	 * center:20dlu:g(2) // center-aligned, 20 dialog units tall, weight=2
	 * c:pref:none       // center-aligned, preferred height, weight=0
	 * p                 // left-aligned (default), preferred height, weight=0
	 * </pre>
	 * 
	 * @param spec
	 * 		the column spec that will be parsed.
	 * @return a ColSpec created by parsing the <code>spec</code> argument.
	 * @throws IllegalArgumentException
	 * 		if the column spec is incorrectly formatted.
	 */
	protected static RowSpec parse(String spec) {
		String[] matches = spec.split("\\s*:\\s*");

		int align = DEFAULT_ALIGN;
		int size = DEFAULT_SIZE;
		int weight = DEFAULT_WEIGHT;

		if (matches.length == 0)
			throw new IllegalArgumentException("Missing row spec");
		else if (matches.length == 1) {
			size = parseSize(matches[0]);
		} else if (matches.length == 2) {
			// Two possible patterns:
			// 1. align:size
			// 2. size:weight
			if (isAlign(matches[0]) && isSize(matches[1])) {
				align = parseAlign(matches[0]);
				size = parseSize(matches[1]);
			} else if (isSize(matches[0]) && isWeight(matches[1])) {
				size = parseSize(matches[0]);
				weight = parseWeight(matches[1]);
			} else {
				throw new IllegalArgumentException("Row spec \"" + spec
						+ "\" does not match spec format.");
			}
		} else if (matches.length == 3) {
			align = parseAlign(matches[0]);
			size = parseSize(matches[1]);
			weight = parseWeight(matches[2]);
		}

		return new RowSpec(align, size, weight);
	}

	// Alignment patterns
	private static final Pattern TOP_ALIGN_PATTERN = Pattern.compile("t|top",
			Pattern.CASE_INSENSITIVE);
	private static final Pattern CENTER_ALIGN_PATTERN = Pattern.compile(
			"c|center", Pattern.CASE_INSENSITIVE);
	private static final Pattern BOTTOM_ALIGN_PATTERN = Pattern.compile(
			"b|bottom", Pattern.CASE_INSENSITIVE);
	private static final Pattern DEFAULT_ALIGN_PATTERN = Pattern.compile(
			"d|def|default", Pattern.CASE_INSENSITIVE);
	private static final Pattern FILL_ALIGN_PATTERN = Pattern.compile("f|fill",
			Pattern.CASE_INSENSITIVE);
	private static final Pattern ANY_ALIGN_PATTERN = Pattern.compile(
			"t|top|c|center|b|bottom|d|def|default|f|fill",
			Pattern.CASE_INSENSITIVE);

	private static boolean isAlign(String align) {
		return ANY_ALIGN_PATTERN.matcher(align).matches();
	}

	static int parseAlign(String align) {
		if (TOP_ALIGN_PATTERN.matcher(align).matches())
			return SWT.TOP;
		else if (CENTER_ALIGN_PATTERN.matcher(align).matches())
			return SWT.CENTER;
		else if (BOTTOM_ALIGN_PATTERN.matcher(align).matches())
			return SWT.BOTTOM;
		else if (DEFAULT_ALIGN_PATTERN.matcher(align).matches())
			return SWT.DEFAULT;
		else if (FILL_ALIGN_PATTERN.matcher(align).matches())
			return SWT.FILL;
		else
			throw new IllegalArgumentException("Invalid alignment: \"" + align
					+ "\"");
	}

	RowSpec copy() {
		return new RowSpec(align, size, weight);
	}
}
